package cn.com.qmth.examcloud.service.examwork.service;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import javax.persistence.criteria.Predicate;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;

import cn.com.qmth.examcloud.common.dto.core.Course;
import cn.com.qmth.examcloud.common.dto.core.Org;
import cn.com.qmth.examcloud.common.dto.core.Student;
import cn.com.qmth.examcloud.common.dto.core.User;
import cn.com.qmth.examcloud.common.dto.core.enums.UserType;
import cn.com.qmth.examcloud.common.util.ErrorMsg;
import cn.com.qmth.examcloud.common.util.excel.ExcelError;
import cn.com.qmth.examcloud.common.util.excel.ExcelReader;
import cn.com.qmth.examcloud.common.util.excel.ExcelReaderHandle;
import cn.com.qmth.examcloud.service.examwork.assembler.ExamStudentAssembler;
import cn.com.qmth.examcloud.service.examwork.dao.ExamRepo;
import cn.com.qmth.examcloud.service.examwork.dao.ExamStudentRepo;
import cn.com.qmth.examcloud.service.examwork.dto.ExamStudentDTO;
import cn.com.qmth.examcloud.service.examwork.entity.Exam;
import cn.com.qmth.examcloud.service.examwork.entity.ExamStudent;
import cn.com.qmth.examcloud.service.examwork.service.rpc.CourseService;
import cn.com.qmth.examcloud.service.examwork.service.rpc.OrgService;
import cn.com.qmth.examcloud.service.examwork.service.rpc.StudentService;

/**
 * 考试学生服务类
 * Created by songyue on 17/1/14.
 */
@Service
public class ExamStudentService {

	protected static final Logger log = LoggerFactory.getLogger(ExamStudentService.class);

    @Autowired
    ExamStudentRepo examStudentRepo;
    @Autowired
    ExamRepo examRepo;
    @Autowired
    ExamStudentAssembler examStudentAssembler;
    @Autowired
    StudentService studentService;
    @Autowired
    CourseService courseService;
    @Autowired
    OrgService orgService;
    
//    @Value("${app.em.photo.path}")
//    private String PHOTO_PATH ;
//    @Value("${app.em.facepp.key}")
//    private String FACEPP_KEY;
//    @Value("${app.em.facepp.secret}")
//    private String FACEPP_SECRET;
    
    /**
     * 获取所有考试学生(分页)
     * @param examCriteria
     * @param pageable
     * @return 
     * @return
     */
    public Page<ExamStudent> getAllExamStudent(ExamStudentDTO examCriteria, Pageable pageable){
//		ExamStudent examStudentSearch = BeanCopierUtil.copyProperties(examCriteria,ExamStudent.class);
//        ExampleMatcher exampleMatcher = ExampleMatcher.matching()
//                .withMatcher("name",contains())
//				.withMatcher("studentCode",contains())
//				.withMatcher("specialtyName",contains()).withIgnoreNullValues();
//        Example<ExamStudent> examExamStudentple = Example.of(examStudentSearch, exampleMatcher);
//		Page<ExamStudent> examStudents = examStudentRepo.findAll(examExamStudentple,pageable);
//        return examStudents;
    	Specification<ExamStudent> specification = getSpecification(examCriteria);
		Page<ExamStudent> examStudents = examStudentRepo.findAll(specification,pageable);
        return examStudents;
    }

    /**
     * 获取所有考试学生
     * @param examCriteria
     * @return 
     * @return
     */
    public  List<ExamStudent> getAllExamStudent(ExamStudentDTO examCriteria){
    	Specification<ExamStudent> specification = getSpecification(examCriteria);
        return examStudentRepo.findAll(specification);
    }

    /**
     * 按ID获取考试学生
     * @param studentId
     * @return
     */
    public ExamStudent getExamStudentById(Long studentId){
        return examStudentRepo.findOne(studentId);
    }

    /**
     * 按考试批次获取学生
     * @param examId
     * @return 
     * @return
     */
    public Page<ExamStudent> getExamStudentByExamId(Long examId,Pageable pageable){
        return examStudentRepo.findByExamId(examId,pageable);
    }

    /**
     * 保存考试学生
     * @param examStudent
     * @return
     */
    public ExamStudent saveExamStudent(ExamStudent examStudent)throws Exception{
    	checkExamStudent(examStudent);
        return saveStudent(examStudent);
    }

    /**
     * 删除考试学生
     * @param examId
     * @return
     */
    public void deleteExamStudent(Long examId){
        examStudentRepo.delete(examId);
    }

    /**
     * 删除所有考试学生
     * @return
     */
    public void deleteAllExamStudent(){
        examStudentRepo.deleteAll();
    }
    
    /**
     * 导入考试学生
     * @return
     */
	public List<ExcelError> importExamStudent(Long examId, InputStream inputStream) {
		log.info("导入考生开始");
		Exam exam = examRepo.findOne(examId);
		List<ExcelError> excelErrors = new ArrayList<ExcelError>();
		if(exam == null){
			ExcelError excelError = new ExcelError(1,"请先创建考试");
			excelErrors.add(excelError);
			return excelErrors;
		}
		List<ExamStudent> examStudents = new ArrayList<ExamStudent>();
		ExcelReader excelReader = new ExcelReader(ExamStudentDTO.class);
        excelErrors = excelReader.reader(inputStream, new ExcelReaderHandle() {
            @Override
            public ExcelError handle(Object obj) {
                ExamStudentDTO dto = (ExamStudentDTO) obj;
                dto.setRootOrgId(exam.getOrgId());
                dto.setExam(exam);
                ExcelError error = importCheck(dto);
                if (error == null) {
                	examStudents.add(examStudentAssembler.toDomain(dto));
                }
                int stu_num = examStudents.size();
                if (stu_num % 5000 == 0) {
                	saveStudentList(examStudents);
					log.info("已导入"+stu_num+"条数据");
                }
                return error;
            }

        });
		int stu_num = examStudents.size();
        saveStudentList(examStudents);
		log.info("已导入"+stu_num+"条数据");
		log.info("导入考生结束");
        return excelErrors;
	}
	

	private void saveStudentList(List<ExamStudent> examStudents) {
		for (ExamStudent examStudent : examStudents) {
			saveStudent(examStudent);
		}
		examStudents.removeAll(examStudents);
	}

	private ExamStudent saveStudent(ExamStudent examStudent){
		if(examStudent.getStudentId() == null){
			Student student = new Student();
			student.setIdentityNumber(examStudent.getIdentityNumber());
			student.setName(examStudent.getName());
			student.setStudentCode(examStudent.getStudentCode());
			student.setRootOrgId(examStudent.getRootOrgId());
			student.setOrgId(examStudent.getOrgId());
			User user = new User();
			user.setOrgId(examStudent.getOrgId());
			user.setEnable(true);
			user.setType(UserType.STUDENT);
			user.setRootOrgId(examStudent.getRootOrgId());
			student.setUser(user);
			student = studentService.addStudent(student);
			examStudent.setStudentId(student.getId());
		}
		examStudent.setFinished(false);
		examStudent.setGraduated(false);
		examStudent.setDegree(false);
		examStudent.setRepair(false);
		return examStudentRepo.save(examStudent);
	}

	public void checkExamStudent(ExamStudent examStudent){
		if(examStudent.getId() != null){
			if(examStudentRepo.checkExamStuById(examStudent.getRootOrgId(),examStudent.getIdentityNumber(),examStudent.getCourseCode(),examStudent.getId()) > 0){
				throw new RuntimeException("该考生已存在");
			}
		}else{
			if(examStudentRepo.checkExamStu(examStudent.getRootOrgId(),examStudent.getIdentityNumber(),examStudent.getCourseCode()) > 0){
				throw new RuntimeException("该考生已存在");
			}
		}
	}
	
	/**
	 * 考生导入验证
	 * @param dto
	 * @return
	 */
	public ExcelError importCheck(ExamStudentDTO dto){
		if(StringUtils.isEmpty(dto.getName())){
			return new ExcelError("姓名不能为空");
		}
		if(StringUtils.isEmpty(dto.getStudentCode())||StringUtils.isEmpty(dto.getIdentityNumber())){
			return new ExcelError("学号或身份证号不能为空");
		}else{
			if(!StringUtils.isEmpty(dto.getStudentCode()) && dto.getStudentCode().length() < 6){
				return new ExcelError("学号至少为6位");
			}
			if(!StringUtils.isEmpty(dto.getIdentityNumber()) && dto.getIdentityNumber().length() < 6){
				return new ExcelError("身份证号至少为6位");
			}
		}
		if(StringUtils.isEmpty(dto.getCourseCode())){
			return new ExcelError("课程代码不能为空");
		}
		if(StringUtils.isEmpty(dto.getOrgCode())){
			return new ExcelError("学习中心代码不能为空");
		}
		Course course = courseService.findByOrgIdAndCode(dto.getRootOrgId(), dto.getCourseCode());
		if(course == null || !course.getEnable()){
			return new ExcelError("课程不存在/被禁用");
		}
		dto.setCourseLevel(course.getLevel().toString());
		dto.setCourseName(course.getName());
		Org org = orgService.findByParentIdAndCode(dto.getRootOrgId(), dto.getOrgCode());
		if(org == null || !org.getEnable()){
			return new ExcelError("学习中心不存在/禁用");
		}
		dto.setOrgId(org.getId());
		dto.setOrgName(org.getName());
		stuCheck(dto);
		return null;
	}

	public void stuCheck(ExamStudentDTO dto){
		ExamStudent examStudent = null;
		if(!StringUtils.isEmpty(dto.getIdentityNumber())) {
			examStudent = examStudentRepo.findFirstByExamId(dto.getExam().getId(), dto.getCourseCode(), dto.getIdentityNumber());
		}else{
			examStudent = examStudentRepo.findFirstByExamId(dto.getExam().getId(), dto.getCourseCode(), dto.getStudentCode());
		}
		if(examStudent != null){
			dto.setId(examStudent.getId());
		}
	}

	/**
	 * 照片检验
	 * @param examId
	 * @return
	 */
	public List<ErrorMsg> photoCheck(Long examId) {
		List<ErrorMsg> msgs = new ArrayList<ErrorMsg>();
		List<ExamStudent> examStudents = examStudentRepo.findByExamId(examId);
		for (ExamStudent examStudent : examStudents) {
			ErrorMsg errorMsg = null;
			if(!StringUtils.isEmpty(examStudent.getStudentCode())){
				
				errorMsg = new ErrorMsg(examStudent.getStudentCode());
			}else if(!StringUtils.isEmpty(examStudent.getIdentityNumber())){
				
				errorMsg =  new ErrorMsg(examStudent.getIdentityNumber());
			}
			if(errorMsg != null){
				msgs.add(errorMsg);
			}
		}
		return msgs;
	}
	
	/**
	 * 生成查询条件
	 * @param examCriteria
	 * @return
	 */
	private Specification<ExamStudent> getSpecification(ExamStudentDTO examCriteria) {
		Specification<ExamStudent> specification = (root, query, cb) -> {
		    List<Predicate> predicates = new ArrayList<>();
		    if(!StringUtils.isEmpty(examCriteria.getRootOrgId())){
		    	predicates.add(cb.equal(root.get("rootOrgId"),examCriteria.getRootOrgId()));
		    }
		    if(!StringUtils.isEmpty(examCriteria.getOrgId())){
		    	predicates.add(cb.equal(root.get("orgId"),examCriteria.getOrgId()));
		    }
		    if(!StringUtils.isEmpty(examCriteria.getOrgName())){
		    	predicates.add(cb.like(root.get("orgName"),"%"+examCriteria.getOrgName()+"%"));
		    }
		    if(!StringUtils.isEmpty(examCriteria.getExamId())){
		    	predicates.add(cb.equal(root.get("exam").get("id"),examCriteria.getExamId()));
		    }
		    if(!StringUtils.isEmpty(examCriteria.getName())){
		    	predicates.add(cb.like(root.get("name"),"%"+examCriteria.getName()+"%"));
		    }
		    if(!StringUtils.isEmpty(examCriteria.getStudentCode())){
		    	predicates.add(cb.like(root.get("studentCode"),"%"+examCriteria.getStudentCode()+"%"));
		    }
			if(!StringUtils.isEmpty(examCriteria.getCourseCode())){
				predicates.add(cb.equal(root.get("courseCode"),examCriteria.getCourseCode()));
			}
			if(!StringUtils.isEmpty(examCriteria.getCourseLevel())){
				predicates.add(cb.equal(root.get("courseLevel"),examCriteria.getCourseLevel()));
			}
			if(!StringUtils.isEmpty(examCriteria.getCourseName())){
				predicates.add(cb.like(root.get("courseName"),"%"+examCriteria.getCourseName()+"%"));
			}
		    if(!StringUtils.isEmpty(examCriteria.getFinished())){
		    	predicates.add(cb.equal(root.get("finished"),examCriteria.getFinished()));
		    }
		    if(!StringUtils.isEmpty(examCriteria.getStudentId())){
		    	predicates.add(cb.equal(root.get("studentId"),examCriteria.getStudentId()));
		    }
			if(!StringUtils.isEmpty(examCriteria.getExamSite())){
				predicates.add(cb.like(root.get("examSite"),"%"+examCriteria.getExamSite()+"%"));
			}
			if(!StringUtils.isEmpty(examCriteria.getIdentityNumber())){
				predicates.add(cb.equal(root.get("identityNumber"),examCriteria.getIdentityNumber()));
			}
			if(!StringUtils.isEmpty(examCriteria.getSpecialtyName())){
				predicates.add(cb.like(root.get("specialtyName"),"%"+examCriteria.getSpecialtyName()+"%"));
			}
			if(!StringUtils.isEmpty(examCriteria.getPaperType())){
				predicates.add(cb.equal(root.get("paperType"),examCriteria.getPaperType()));
			}
			if(!StringUtils.isEmpty(examCriteria.getDegree())){
				predicates.add(cb.equal(root.get("degree"),examCriteria.getDegree()));
			}
			if(!StringUtils.isEmpty(examCriteria.getRepair())){
				predicates.add(cb.equal(root.get("repair"),examCriteria.getRepair()));
			}
			if(!StringUtils.isEmpty(examCriteria.getGraduated())){
				predicates.add(cb.equal(root.get("graduated"),examCriteria.getGraduated()));
			}
			if(!StringUtils.isEmpty(examCriteria.getGrade())){
				predicates.add(cb.like(root.get("grade"),"%"+examCriteria.getGrade()+"%"));
			}
			if(!StringUtils.isEmpty(examCriteria.getExamNumber())){
				predicates.add(cb.like(root.get("examNumber"),"%"+examCriteria.getExamNumber()+"%"));
			}
		    return cb.and(predicates.toArray(new Predicate[predicates.size()]));
		};
		return specification;
	}
}
